/*
 * LICTiA2FController.cpp
 *
 *  Created on: 2020/06/16
 *      Author: Tsuyoshi Anazawa
 *
 *     Outline: LICTiA2Fのドア制御プログラム
 *
 */
#include <cnoid/SimpleController>
#include <cnoid/Joystick>
#include <cnoid/RangeSensor>
#include <cnoid/ConnectionSet>
#include <fmt/format.h>
#include <iostream>
#include <sstream>

using namespace std;
using namespace cnoid;
using fmt::format;

class LICTiA2FController : public SimpleController
{
	// Door格納変数
	vector<Link*> door;
	// push/pull型のDoorの数
	static const int DOORNUM = 24;
	static const int SLIDEDOORNUM = 4;
	// Doorの名称
	string doorNames[DOORNUM] = {/* 2F階段,       女子WC,           男子WC */
									"Stair_door", "Woman_wc_door", "Man_wc_door",
								 /*	プロジェクト4, プロジェクト3, プロジェクト2 */
									"Pr4_door_0", "Pr3_door_0", "Pr2_door_0",
								 /* プロジェクト1, UPS左側,           UPS右側 */
									"Pr1_door_0", "UPS_left_door", "UPS_right_door",
								 /*	収納2左側,              収納2右側,             荷揚 */
									"Storage2_left_door",  "Storage2_right_door", "Outside_door",
								 /*	カンファレンス2, カンファレンス1, カンファレンス0右側 */
									"Conf2_door",   "Conf1_door",   "Conf0_right_door",
								 /* カンファレンス0左側, 電気室右側,      電気室左側 */
									"Conf0_left_door", "Er_right_door", "Er_left_door",
								 /* 給気ファン室,  MDF室 */
									"Saf_door_0", "MDF_door",
								 /* 添字が20以降は Slide door  */
								 /* 収納1左,               収納1右,     エレベータ外側左 */
									"Storage1_left_door", "Storage1_right_door", "ELV_DOOR_OUTSIDE_L",
								 /* エレベータ外側右,      風除室外側左 */
									"ELV_DOOR_OUTSIDE_R" };
	double slideDistance[SLIDEDOORNUM] = {/* エレベータ外側左,	エレベータ外側右,		収納1左側 */
											 0.445,				0.4491,				0.6,
										  /* 収納1右側 */
											 0.6 };
	// Body情報格納変数
	Body* body;
	// アクチュエーションモード
	Link::ActuationMode actuationMode;
	double qprev[DOORNUM];
	double qref[DOORNUM];
	double dt;
	// Range sensor情報
	RangeSensorPtr rangeSensor[2];
	std::shared_ptr<const RangeSensor::RangeData> prevRangeData[2];
	char id[13];
	string deviceName;
	double dis[1];	// 0:エレベーター
	bool openFlg;
	// 現在のスコープ限定の接続
	ScopedConnectionSet rangeConnections;

public:
	virtual bool initialize(SimpleControllerIO* io) override
	{
		ostream& os = io->os();
		// Bodyの取得
		body = io->body();

		// アクチュエーションモードの初期化
		actuationMode = Link::JOINT_TORQUE;
		// コントローラオプションの取得
		string option = io->optionString();

		if(!option.empty()){
			// コントローラオプションが空の場合
			if(option == "velocity" || option == "position"){
				// velocity または positionの場合
				actuationMode = Link::JOINT_VELOCITY;
			} else if(option == "torque"){
				// torqueの場合
				actuationMode = Link::JOINT_TORQUE;
			} else{
				// 上記以外の場合
				os << format("Warning unknown option \"{}\".", option) << endl;
			}
		}

		// push/pullのDoor
		for(int i = 0; i < DOORNUM; ++i){
			// 配列にドアリンクを格納
			door.push_back(io->body()->link(doorNames[i]));

			if(!door[i]){
				// リンクが取得できない場合
				os << format("{0} of {1} is not found.", doorNames[i], body->name()) << endl;
				return false;
			}

			// リンクのアクチュエーションモードの設定
			door[i]->setActuationMode(actuationMode);
			// ドアへの出力を有効化
			io->enableIO(door[i]);
			qref[i] = qprev[i] = door[i]->q();
		}

		rangeConnections.disconnect();
		for(int i = 0; i < body->devices<RangeSensor>().size(); ++i){
			// レンジセンサ番号を設定
			sprintf(id, "%d", i);
			// レンジセンサ名を格納
			deviceName = "RangeSensor_" + (string)id;
			// レンジセンサを取得
			rangeSensor[i] = body->findDevice<RangeSensor>(deviceName);

			if(!rangeSensor[i]){
				// 該当するレンジセンサが存在しない場合
				os << format("{} is not found.", deviceName) << endl;
				return false;
			}

			// レンジセンサからの入力を有効化
			io->enableInput(rangeSensor[i]);
			initializeRangeSensor();
			// センサの状態が変わった場合、onRangeSensorStateChanged()を呼び出す
			rangeConnections.add(rangeSensor[i]->sigStateChanged().connect(
					[&](){ onRangeSensorStateChanged(); }));
		}

		dt = io->timeStep();
		openFlg = false;

		return true;
	}

	virtual bool control() override
	{
		for(int i = 0; i < DOORNUM; ++i){
			double u;

			if(dt == 0.0){
				return false;
			}

			double dq_current = (door[i]->q() - qprev[i]) / dt;
			double dq_ref = 0.0;

			if(i < (DOORNUM - SLIDEDOORNUM)){
				// PD制御定数
				const double Kp = 25.0;
				const double Kd = 25.0;

				if(door[i]->q() < 0.0 or door[i]->q() > round(M_PI * 100) / 200){
					qref[i] += dt * 0.05;
				} else if(door[i]->q() > 0.1 and door[i]->q() < round(M_PI * 100) / 200 - 0.1){
					qref[i] += dt * (-0.05);
				} else {
					qref[i] = 0.0;
				}

				u = Kp * (qref[i] - door[i]->q()) + Kd * (dq_ref - dq_current);

			} else if(i >= (DOORNUM - SLIDEDOORNUM)){
				// PD制御定数
				double Kp = 100.0;
				double Kd = 350.0;
				// slide式のドアの場合
				if(i <= 21){
					if(door[i]->q() < 0.0 or door[i]->q() >= slideDistance[2]){
						qref[i] += 0.05 * dt;
					} else if(door[i]->q() > 0.1 and door[i]->q() < slideDistance[0] - 0.1){
						qref[i] += (-0.05) * dt;
					} else {
						qref[i] = 0.0;
					}
					Kp = 200.0;
					Kd = 250.0;

					u = Kp * (qref[i] - door[i]->q()) + Kd * (dq_ref - dq_current);
				} else if(i >= 22 and dis[0] < 1.7){
					// 自動ドア かつ レンジセンサで取得した距離 < 1.7 の場合
					if(!openFlg){
						// 開閉フラグがfalseの場合
						u = Kp * qref[i] + Kd * (dq_ref - dq_current);
					} else {
						// 上記以外の場合
						if(door[i]->q() >= slideDistance[i - (DOORNUM - SLIDEDOORNUM)] - 0.1){
							qref[i] = slideDistance[i - (DOORNUM - SLIDEDOORNUM)] - 0.1;
						} else {
							qref[i] += 0.5 * dt;
						}
						u = Kp * (qref[i] - door[i]->q()) + Kd * (dq_ref - dq_current);
					}
				} else if(i >= 22 and dis[0] >= 1.7){
					// 自動ドア かつ レンジセンサで取得した距離 >= 1.7 の場合
					if(openFlg){
						// 開閉フラグがfalseの場合
						u = Kp * qref[i] + Kd * (dq_ref - dq_current);
					} else {
						// 上記以外の場合
						if(door[i]->q() <= 0.1){
							qref[i] = 0.0;
						} else {
							if(qref[i] > 0.0){
								qref[i] = 0.0;
							} else {
								qref[i] += (-0.5) * dt;
							}
						}
						u = Kp * (qref[i] - door[i]->q()) + Kd * (dq_ref - dq_current);
					}
				}
			} else{
				// 上記以外の場合
				u = 0.0;
			}
			door[i]->u() = u;
			qprev[i] = door[i]->q();
		}
		return true;
	}

	virtual void stopSimulation()
	{
		for(auto sensor : rangeSensor){
			sensor.reset();
		}
	}

	void initializeRangeSensor()
	{
		dis[0] = 100.0;
	}

	void onRangeSensorStateChanged()
	{
		// レンジセンサの数分処理を繰り返す
		for(int i = 0; i < body->devices<RangeSensor>().size(); ++i){
			if(rangeSensor[i]->sharedRangeData() != prevRangeData[i]){
				// レンジセンサの値が前回値と異なる場合
				// 前回値を現在値で更新
				prevRangeData[i] = rangeSensor[i]->sharedRangeData();
				// レンジセンサの値を取得
				const RangeSensor::RangeData& src = rangeSensor[i]->constRangeData();

				// 取得したレンジセンサの値分処理を繰り返す
				for(size_t j = 0; j < src.size(); ++j){
					if(i >= 0 and i <= 1){
						// 外側の自動ドアの場合
						if(src[j] < 1.7){
							// レンジセンサの値 < 1.7 の場合
							dis[0] = src[j];
							openFlg = true;
							break;
						} else {
							// 上記以外の場合
							dis[0] = src[j];
							openFlg = false;
						}
					}
				}
			}
			if(openFlg){
				// 開閉フラグ = true の場合
				break;
			}
		}
	}
};

CNOID_IMPLEMENT_SIMPLE_CONTROLLER_FACTORY(LICTiA2FController)
